#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/time.h>
#include "ras-logger.h"
#include "trigger.h"
#include <string.h>

#define READ    0
#define WRITE   1

static int child_done, alarm_done;
static char *trigger_dir;

static void child_handler(int sig)
{
    child_done = 1;
}

static void alarm_handler(int sig)
{
    alarm_done = 1;
}

void run_trigger(struct event_trigger *t, char *argv[], char **env,
                        const char* reporter, char *msg)
{
        pid_t child;
        char *path, err[256] = {0}, *trigger = t->path;
        int status, pipe_stdout[2], pipe_stderr[2], timeout = t->timeout;
        ssize_t byte = 0;

        log(TERM, LOG_INFO, "Running trigger `%s' (reporter: %s)\n", trigger, reporter);

	if (asprintf(&path, "%s/%s", trigger_dir, trigger) < 0)
		return;

        if (pipe(pipe_stdout) == -1)
                exit(EXIT_FAILURE);

        if (pipe(pipe_stderr) == -1)
                exit(EXIT_FAILURE);

        child = fork();
        if (child < 0) {
                log(TERM, LOG_ERR, "Cannot create process for trigger");
                return;
        } else if (child == 0) {
                close(pipe_stdout[READ]);
                close(pipe_stderr[READ]);
                dup2(pipe_stdout[WRITE], 1);
                dup2(pipe_stderr[WRITE], 2);
                close(pipe_stdout[WRITE]);
                close(pipe_stderr[WRITE]);

                execve(path, argv, env);
                exit(EXIT_FAILURE);
        }

        signal(SIGCHLD, child_handler);

        close(pipe_stdout[WRITE]);
        close(pipe_stderr[WRITE]);

        if (timeout) {
                signal(SIGALRM, alarm_handler);
                alarm(timeout);
        }

        pause();

        if (child_done) {
                if (waitpid(child, &status, WNOHANG) == child){
                        if (WIFEXITED(status) && WEXITSTATUS(status))
                                log(TERM, LOG_INFO, "Trigger %s exited with status %d\n",
                                        trigger, WEXITSTATUS(status));
                        else if (WIFSIGNALED(status))
                                log(TERM, LOG_INFO, "Trigger %s killed by signal %d\n",
                                        trigger, WTERMSIG(status));
                }
                alarm(0);
        } else if (alarm_done) {
                log(TERM, LOG_ERR, "Trigger timeout, kill it\n");
                kill(child, SIGKILL);
        }
        signal(SIGCHLD, SIG_DFL);
        signal(SIGALRM, SIG_DFL);

        byte = read(pipe_stderr[READ], err, 256);
        if (byte > 0)
                log(TERM, LOG_ERR, "Trigger stderr: %s\n", err);
        else if (byte < 0)
                log(TERM, LOG_ERR, "Trigger error : %s\n", strerror(byte));

        byte = read(pipe_stdout[READ], msg, 4096);
        if (byte < 0)
                log(TERM, LOG_ERR, "Trigger error : %s\n", strerror(byte));

        close(pipe_stdout[READ]);
        close(pipe_stderr[READ]);
}

int trigger_check(char *s)
{
        char *name;
        int rc;

        if (trigger_dir) {
                if (asprintf(&name, "%s/%s", trigger_dir, s) < 0)
                        return -1;
        } else
                name = s;

        rc = access(name, R_OK|X_OK);

        if (trigger_dir)
                free(name);

        return rc;
}

static struct event_trigger *event_triggers[] = {
	&mc_ce_trigger,
        &mc_ue_trigger,
#ifdef HAVE_MCE
        &mce_ce_trigger,
        &mce_de_trigger,
        &mce_ue_trigger,
#endif
};

void trigger_setup(void)
{
        int i, j;
        struct event_trigger *trigger;
        char *s, timeout_env[30];

        trigger_dir = getenv("TRIGGER_DIR");

        for (i = 0; i < ARRAY_SIZE(event_triggers); i++) {
                trigger = event_triggers[i];

                s = getenv(trigger->env);
                if (!s || !strcmp(s, ""))
		        continue;

                trigger->path = s;
                if (trigger_check(s) < 0)
                        log(SYSLOG, LOG_ERR, "Cannot access trigger `%s`\n", s);
                else
                        log(SYSLOG, LOG_NOTICE, "Setup %s trigger `%s`\n",
                                                        trigger->event_name, s);

                sprintf(timeout_env, "%s_TIMEOUT", trigger->env);

                trigger->timeout = 1;
                s = getenv(timeout_env);
                if (!s || !strcmp(s, "")) {
                        log(SYSLOG, LOG_NOTICE,
                                "Setup %s trigger default timeout 1s",
                                        trigger->event_name);
		        continue;
                }

                j = atoi(s);
                if (j < 0)
                        log(SYSLOG, LOG_ERR,
                                "Invalid %s trigger timeout `%d`"
                                "use default value: 1s\n",
                                        trigger->event_name, j);
                else if (j == 0) {
                        log(SYSLOG, LOG_NOTICE,
                                "%s trigger no timeout\n", trigger->event_name);
                        trigger->timeout = 0;
                } else {
                        log(SYSLOG, LOG_NOTICE,
                                "Setup %s trigger timeout `%d`s\n",
                                trigger->event_name, j);
                        trigger->timeout = j;
                }
        }
}
